@@ -16,8 +16,8 @@
 //
 //
 
-#ifndef GRPC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
-#define GRPC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
+#ifndef GRPC_SRC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
+#define GRPC_SRC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
 
 #include <grpc/support/port_platform.h>
 
@@ -44,4 +44,4 @@ inline bool IncrementIfNonzero(std::atomic<T>* p) {
 
 }  // namespace grpc_core
 
-#endif  // GRPC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
+#endif  // GRPC_SRC_CORE_LIB_GPRPP_ATOMIC_UTILS_H
